/*
 * Copyright (c) 2000 Shane Nay
 * Portions heavily modified by Greg Haerr
 *
 * Win32 API upper level font selection routines
 */

/* Font routines added by Shane for fltk support.
   Copyright Shane Nay 2000.
   Feel free to borrow, steal, hack for whatever purpose you see
   fit.  There is no guarantee of any form.  You use this software
   at your own risk.  By using, compiling, or simply procuring this
   software you must agree to not hold me personally responsible for
   any damage that this software may cause to your computer.  In
   addition you cannot hold any company that I work for, or represent
   at this time, in the past, or in the future for such damages.
*/
#include "windows.h"
#include "wintern.h"
#include <stdlib.h>

HFONT WINAPI
CreateFont(int nHeight, int nWidth, int nEscapement, int nOrientation,
	int fnWeight, DWORD fdwItalic, DWORD fdwUnderline, DWORD fdwStrikeOut,
	DWORD fdwCharSet,DWORD fdwOutputPrecision,DWORD fdwClipPrecision,
	DWORD fdwQuality, DWORD fdwPitchAndFamily, LPCSTR lpszFace)
{
	FONTOBJ * 	hfont;
	int 		i;
	int		fontht;
	int		fontno = 0;
	FONTINFO	fontinfo;
	SCREENINFO 	scrinfo;

	/* create a gdi font object*/
	hfont = ItemNew(FONTOBJ);
	if(!hfont)
		return NULL;
	hfont->hdr.type = OBJ_FONT;
	hfont->hdr.stockobj = FALSE;

	/* get screen information*/
	GdGetScreenInfo(&scrdev, &scrinfo);

	/* find font closest in height*/
	nHeight = abs(nHeight);
	fontht = COORD_MAX;
	for(i=0; i < scrinfo.fonts; ++i) {
		GdGetFontInfo(&scrdev, i, &fontinfo);
		if(fontht > abs(nHeight-fontinfo.height)) { 
			fontno = i;
			fontht = abs(nHeight-fontinfo.height);
		}
	}
	hfont->fontno = fontno;
	return (HFONT)hfont;
}

BOOL WINAPI
GetTextMetrics(HDC hdc, LPTEXTMETRIC lptm)
{
	FONTINFO 	fi;

	if(!hdc)
		return FALSE;

	GdGetFontInfo(hdc->psd, hdc->font->fontno, &fi);

	/* FIXME many items are guessed for the time being*/
	lptm->tmHeight = fi.height;
	lptm->tmAscent = fi.height - fi.baseline;
	lptm->tmDescent= fi.baseline;
	lptm->tmInternalLeading = 0;
	lptm->tmExternalLeading = 0;
	lptm->tmAveCharWidth = fi.widths['x'];
	lptm->tmMaxCharWidth = fi.maxwidth;
	lptm->tmWeight = FW_NORMAL;
	lptm->tmOverhang = 0;
	lptm->tmDigitizedAspectX = fi.maxwidth;
	lptm->tmDigitizedAspectY = fi.height;
	lptm->tmFirstChar = 32;
	lptm->tmLastChar = 255;
	lptm->tmDefaultChar = '?';
	lptm->tmBreakChar = 0;
	lptm->tmItalic = 0;
	lptm->tmUnderlined = 0;
	lptm->tmStruckOut = 0;
	/* note that win32 has the TMPF_FIXED_PITCH flags REVERSED...*/
	lptm->tmPitchAndFamily = fi.fixed?
			FF_DONTCARE: (FF_DONTCARE | TMPF_FIXED_PITCH);
	lptm->tmCharSet = OEM_CHARSET;
	return TRUE;
}

BOOL WINAPI
GetCharWidth(HDC hdc, UINT iFirstChar, UINT iLastChar, LPINT lpBuffer)
{
	int 		i;
	int		j = 0;
	FONTINFO	fi;

	if(!hdc || iLastChar < iFirstChar || iLastChar > 255)
		return FALSE;

	GdGetFontInfo(hdc->psd,hdc->font->fontno, &fi);
	for(i=iFirstChar; i <= iLastChar; ++i)
		lpBuffer[j++] = fi.widths[i];

	return TRUE;
}
